# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2018
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.app import route
from pixiedust.utils import Logger

@Logger()
class RowCount(object):
    @route(widget="pdRowCount")
    def chart_option_rowcount_widget(self, optid, count):
        self.rowCount = count or 100
        return """
<script>
    var rowCountStatus{{optid}}{{prefix}} = true
    function rowCount{{optid}}{{prefix}}(input) {
        var val = $(input).val()
        var invalid = !val || isNaN(val) || Math.floor(Number(val)) < 1
        if (rowCountStatus{{optid}}{{prefix}} && invalid) {
            rowCountStatus{{optid}}{{prefix}} = false
            chartOptionsOKStatus{{prefix}}(true)
        } else if (!rowCountStatus{{optid}}{{prefix}} && !invalid) {
            chartOptionsOKStatus{{prefix}}(false)
            rowCountStatus{{optid}}{{prefix}} = true
            $(input).trigger('click')
        }
    }
</script>
<div class="form-group">
    <label class="field"># of Rows to Display:</label>
    <input id="chartoption{{optid}}{{prefix}}" class="form-control" name="{{optid}}" type="number" min="1" max="10000" value="{{this.rowCount}}"
          pd_script="self.options_callback('{{optid}}','$val(chartoption{{optid}}{{prefix}})')" onkeyup="rowCount{{optid}}{{prefix}}(this)">
</div>
"""



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2018
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.app import *
from pixiedust.utils import Logger
from pixiedust.display.chart.options.optionsShell import *
from pixiedust.display.chart.options.components.TableValueSelector import *
from pixiedust.display.chart.options.components.AggregationSelector import *
from pixiedust.display.chart.options.components.RowCount import *

@PixieApp
@Logger()
class TableOptions(OptionsShell, TableValueSelector, AggregationSelector, RowCount):
    def setup(self):
        OptionsShell.setup(self)
        self.chart_options.append({
            "optid": "tablevalue",
            "classname": "no_loading_msg",
            "tableFields": lambda: self.options.get("tableFields") or "",
            "widget": "pdTableValueSelector"
        })

        self.chart_options.append({
            "optid": "rowCount",
            "classname": "field-width-50 no_loading_msg",
            "count": lambda: self.options.get("rowCount") or 100,
            "widget": "pdRowCount"
        })



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
from abc import abstractmethod, ABCMeta
from six import with_metaclass
import pandas
from pixiedust.utils.dataFrameAdapter import PandasDataFrameAdapter

__all__ = ['StreamingDataAdapter'] 

class StreamingDataAdapter(with_metaclass(ABCMeta)):
    def __init__(self):
        self.channels = []
        
    def getNextData(self):
        nextData = self.doGetNextData()
        if nextData is not None and hasattr(self, "channels"):
            for channel in self.channels:
                channel.processNextData(nextData)
        return nextData

    def getMetadata(self):
        return {}

    def accept(self, handlerId):
        return True

    defaultValues = {
        "getFieldNames": lambda expandNested=False: [],
        "schema": PandasDataFrameAdapter(pandas.DataFrame()).schema,
        "getWorkingPandasDataFrame": lambda xFields, yFields, extraFields=[], aggregation=None, maxRows = 100, filterOptions={}, isTableRenderer=False: pandas.DataFrame()
    }

    def getDisplayDataHandler(self, options, entity):
        from pixiedust.display.datahandler.baseDataHandler import BaseDataHandler
        this = self
        class StreamingDisplayDataHandler(BaseDataHandler):
            def __init__(self, options, entity):
                super(StreamingDisplayDataHandler, self).__init__(options, entity)
                self.isStreaming = True
            def __getattr__(self, name):
                if hasattr(this, name):
                    return self.this.__getattribute__(name)
                elif name in StreamingDataAdapter.defaultValues:
                    return StreamingDataAdapter.defaultValues[name]
                raise AttributeError("{0} attribute not found".format(name))
            def accept(self, handlerId):
                return this.accept(handlerId)
            def add_numerical_column(self):
                pass
        return StreamingDisplayDataHandler(options, entity)

    @abstractmethod
    def doGetNextData(self):
        """Return the next batch of data from the underlying stream. 
        Accepted return values are:
        1. (x,y): tuple of list/numpy arrays representing the x and y axis
        2. pandas dataframe
        3. y: list/numpy array representing the y axis. In this case, the x axis is automatically created
        4. pandas serie: similar to #3
        5. json
        6. geojson
        7. url with supported payload (json/geojson)
        """
        pass

    def getStreamingChannel(self, processfn, initialData = None):
        channel = StreamingChannel(processfn, initialData)
        if not hasattr(self, "channels"):
            self.channels = []
        self.channels.append(channel)
        return channel
    
class StreamingChannel(StreamingDataAdapter):
    def __init__(self, processfn, initialData):
        super(StreamingChannel,self).__init__()
        self.processfn = processfn
        self.processedData = None
        self.accumulator = initialData

    def processNextData(self, nextData):
        newProcessedData, self.accumulator = self.processfn(self.accumulator, nextData)
        #merge
        self.processedData = newProcessedData if self.processedData is None else self.processedData + newProcessedData
        
    def doGetNextData(self):
        nextData = self.processedData
        self.processedData = None
        return nextData


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
from pixiedust.display.streaming import *
from six import string_types, iteritems, integer_types
import json
import os
import ssl
from kafka import KafkaConsumer
from kafka.errors import KafkaError

class MessagehubStreamingAdapter(StreamingDataAdapter):
    def __init__(self, topic, username, password, prod=True):
        # Create a new context using system defaults, disable all but TLS1.2
        context = ssl.create_default_context()
        context.options &= ssl.OP_NO_TLSv1
        context.options &= ssl.OP_NO_TLSv1_1
        conf = {
            'client_id': 'pixieapp.client.id',
            'group_id': 'pixieapp.group',
            'sasl_mechanism': 'PLAIN',
            'security_protocol': 'SASL_SSL',
            'ssl_context': context,
            "bootstrap_servers": [ "kafka0{}-{}.messagehub.services.us-south.bluemix.net:9093".format(i, "prod01" if prod else "stage1") for i in range(1,6)],
            "sasl_plain_username": username,
            "sasl_plain_password": password,
            "auto_offset_reset":"latest"
        }
        self.consumer = KafkaConsumer(**conf)
        self.consumer.subscribe([topic])
        self.schema = {}
        self.sampleDocCount = 0
        
    def close(self):
        self.consumer.unsubscribe()
        self.consumer.close() 
        
    def tryCast(self, value, t):
        def _innerTryCast(value, t):
            try:
                return t(value)
            except:
                return None

        if isinstance(t, tuple):
            for a in t:
                ret = _innerTryCast(value, a)
                if ret is not None:
                    return ret
            return None
        
        return _innerTryCast(value, t)
        
    def inferType(self, value):
        if isinstance(value, string_types):
            value = self.tryCast(value, integer_types) or self.tryCast(value, float) or value
        return "integer" if value.__class__==int else "float" if value.__class__ == float else "string"
        
    def inferSchema(self, eventJSON):
        if self.sampleDocCount > 20:
            return
        for key,value in iteritems(eventJSON):
            if not key in self.schema:
                self.schema[key] = self.inferType(value)
        self.sampleDocCount = self.sampleDocCount + 1 
    
    def doGetNextData(self):
        msgs = []
        msg = self.consumer.poll(1000, max_records=10)
        if msg is not None:
            for topicPartition,records in iteritems(msg):
                for record in records:
                    if record.value is not None:                    
                        jsonValue = json.loads(record.value.decode('utf-8'))
                        self.inferSchema(jsonValue)
                        msgs.append(jsonValue)
        return msgs
    
    def close(self):
        self.consumer.close()


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
from pixiedust.display.app import *
from pixiedust.apps.gateway import BaseGatewayApp
from pixiedust.utils import Logger
from pixiedust.utils.shellAccess import ShellAccess
from pixiedust.utils.userPreferences import setUserPreference
from IPython.utils.io import capture_output
import requests

@PixieApp
@Logger()
class ShareChartApp(BaseGatewayApp):
    def setup(self):
        BaseGatewayApp.setup(self)
        self.title = "Share Chart and make it accessible on the web"
        self.tab_definitions = [{
            "title": "Basic Shared Options",
            "id": "options",
            "name": "Options",
            "contents": lambda: self.renderTemplate("shareBasicOptions.html")
        }
        # ,{
        #     "title": "Access Control",
        #     "id": "permissions",
        #     "name": "Permissions",
        #     "contents": lambda: self.renderTemplate("sharePermissionsOptions.html")
        # },{
        #     "title": "Chart Refresh Options",
        #     "id": "refresh",
        #     "name": "Refresh",
        #     "contents": lambda: self.renderTemplate("shareRefreshOptions.html")
        # }
        ]
        self.gateway_buttons = [{
            "title": "Share",
            "options": ["server", "description"]
        }]

    def update_command(self, command, key, value):
        has_value = value is not None and value != ''
        pattern = ("" if has_value else ",")+"\\s*" + key + "\\s*=\\s*'((\\\\'|[^'])*)'"
        m = re.search(pattern, str(command), re.IGNORECASE)
        ret_command = command
        if m is not None:
            ret_command = command.replace(m.group(0), key+"='"+value+"'" if has_value else "")
        elif has_value:
            k = command.rfind(")")
            ret_command = command[:k]
            ret_command += ","+key+"='"+value+"'"
            ret_command += command[k:]
        return ret_command

    @route(gateway_server="*")
    def shareIt(self, gateway_server, gateway_description):
        self.server = gateway_server.strip('/')
        setUserPreference("pixie_gateway_server", gateway_server)
        rendererid = ''
        with capture_output() as buf:
            try:
                command = self.parent_command
                #add any options from the current cell metadata
                if self.cell_metadata is not None and 'pixiedust' in self.cell_metadata and 'displayParams' in self.cell_metadata['pixiedust']:
                    for key,value in iteritems(self.cell_metadata['pixiedust']['displayParams']):
                        command = self.update_command(command, key, value)
                        if key == 'rendererId':
                            rendererid = value
                command = self.update_command(command, "nostore_figureOnly", "true")
                command = self.update_command(command, "chart_share", "true")
                command = self.update_command(command, "runInDialog", "false")
                self.debug("Share command: {}".format(command))
                sys.modules['pixiedust.display'].pixiedust_display_callerText = command
                for key in ShellAccess:
                    locals()[key] = ShellAccess[key]
                eval(command, globals(), locals())
            finally:
                del sys.modules['pixiedust.display'].pixiedust_display_callerText

        payload = {
            "chart": "\n".join([output._repr_html_() for output in buf.outputs]),
            "description": gateway_description,
            "rendererId": rendererid
        }
        
        try:
            share_url = "{}/chart".format(self.server)
            response = requests.post(share_url, json = payload)
            if response.status_code == requests.codes.ok:
                self.chart_model = response.json()
                return """
    <style type="text/css">
    .share{
        font-size: larger;
        margin-left: 30px;
    }
    .share h2 {
        padding-bottom: 15px;
    }
    .share h2:not(:first-child) {
        margin-top: 50px;
    }
    .publish .summary{
        font-size: xx-large;
        text-align: center;
    }
    .share textarea{
        font-family: monospace;
        resize: none;
    }
    </style>
    <div class="share">
        <div class="summary">
            <h2>Chart Successfully shared</h2>
            <div>
                <a href="{{this.server}}/chart/{{this.chart_model['CHARTID']}}" target="blank">
                    {{this.server}}/chart/{{this.chart_model['CHARTID']}}
                </a>
            </div>
            <h2>Embed the chart into your web app</h2>
            <div>
                <textarea class="form-control" rows="4"  readonly>
&lt;object type="text/html" data="{{this.server}}/embed/{{this.chart_model['CHARTID']}}/600/400" width="600" height="400"&gt;
    &lt;a href="{{this.server}}/embed/{{this.chart_model['CHARTID']}}"&gt;View Chart&lt;/a&gt;
&lt;/object&gt;
                </textarea>
            </div>
        </div>
    </div>
                """
            
            return "<div>An Error occured while sharing this chart: {}".format(response.text)
        except Exception as ex:
            return "<div>Unexcepted error, Please check that PixieGateway server {} is reachable: {}".format(self.server, ex)



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2018
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.chart.renderers import PixiedustRenderer
from ..baseChartDisplay import BaseChartDisplay
from six import with_metaclass
from abc import ABCMeta

import pixiedust

myLogger = pixiedust.getLogger(__name__)

@PixiedustRenderer(rendererId="mapbox")
class MapBoxBaseDisplay(with_metaclass(ABCMeta, BaseChartDisplay)):
    def get_options_dialog_pixieapp(self):
            """
            Return the fully qualified path to a PixieApp used to display the dialog options
            PixieApp must inherit from pixiedust.display.chart.options.baseOptions.BaseOptions
            """
            return "pixiedust.display.chart.renderers.mapbox.mapboxOptions.MapboxOptions"

    def get_numbins_options_pixieapp(self):
        """
        Return the fully qualified path to a PixieApp used to display a dialog for options
        controlling the break down of bin clustering.
        PixieApp must inherit from pixiedust.display.chart.options.baseOptions.BaseOptions
        """
        return "pixiedust.display.chart.renderers.mapbox.mapboxOptions.NumBinsOptions"

    def getChartOptions(self):
        return [ 
            {
                'name': 'legend',
                'description': 'Show legend',
                'metadata': {
                    'type': 'checkbox',
                    'default': "true"
                }
            },
            {
                'name': 'chartsize',
                'description': 'Map Size',
                'metadata': {
                        'type': 'slider',
                        'max': 100,
                        'min': 50,
                        'default': 90
                }
            },
            {
                'name': 'coloropacity',
                'description': 'Opacity',
                'metadata': {
                        'type': 'slider',
                        'max': 100,
                        'min': 0,
                        'default': 80
                }
            },
            {
                'name': 'kind',
                'description': 'Style',
                'metadata': {
                    'type': 'dropdown',
                    'values': ['simple','simple-cluster','choropleth','choropleth-cluster','densitymap'],
                    'default': 'simple'
                }
            },
            {
                'name': 'colorrampname',
                'description': 'Color Ramp',
                'metadata': {
                    'type': 'dropdown',
                    'values': [
                        'Yellow to Blue',
                        'Light to Dark Red',
                        'Grayscale',
                        'Orange to Purple',
                        'Green to Purple',
                        'Custom Monochrome Light',
                        'Custom Monochrome Dark',
                        'Custom Color Range'
                    ],
                    'default': 'Yellow to Blue'
                }
            },
            {
                'name': 'basemap', 
                'metadata': {
                    'type': 'dropdown', 
                    'values': ['light-v9', 'satellite-v9', 'dark-v9','outdoors-v9'], 
                    'default': 'light-v9'
                }
            },
            {
                'name': 'numbins',
                'description': 'Number of Bins',
                'metadata': {
                    'type': 'slider',
                    'max': 20,
                    'min': 2,
                    'default': 5
                }
            },
            {
                'name': 'binoptions',
                'description': 'Bin Options',
                'metadata': {
                    'type': 'button',
                    'pd_app': self.get_numbins_options_pixieapp()
                }
            }
        ]



from setuptools import setup, find_packages
setup(name='pixiedust',
      version='1.1.15',
      description='Productivity library for Jupyter Notebook',
      url='https://github.com/pixiedust/pixiedust',
      install_requires=['mpld3', 'lxml', 'geojson', 'astunparse', 'markdown', 'colour', 'requests'],
      author='David Taieb',
      author_email='david_taieb@us.ibm.com',
      license='Apache 2.0',
      packages=find_packages(exclude=('tests', 'tests.*')),
      include_package_data=True,
      zip_safe=False,
      entry_points={
          'console_scripts': [
              'jupyter-pixiedust = install.pixiedustapp:main'
          ]
      }
     )



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.chart.renderers import PixiedustRenderer
from .mapBoxBaseDisplay import MapBoxBaseDisplay
from pixiedust.utils import cache
from pixiedust.utils import Logger
from pixiedust.utils.shellAccess import ShellAccess
from pixiedust.display.streamingDisplay import *
from colour import Color
import json
import numpy
import geojson
import uuid
import requests
import colorsys

def defaultJSONEncoding(o):
    if isinstance(o, numpy.integer): 
        return int(o)
    raise TypeError

@PixiedustRenderer(id="mapView")
@Logger()
class MapViewDisplay(MapBoxBaseDisplay):
    def isMap(self, handlerId):
        return True

    def supportsAggregation(self, handlerId):
        return False

    def supportsLegend(self, handlerId):
        return True

    def getPreferredDefaultKeyFieldCount(self, handlerId):
        return 2

    def getPreferredDefaultValueFieldCount(self, handlerId):
        return 1

    def canRenderChart(self):
        keyFields = self.getKeyFields()
        if len(self.getFieldNames()) == 0 or (keyFields is not None and len(keyFields) > 0) or len(self._getDefaultKeyFields()) > 0:
            return (True, None)
        else:
            return (False, "No location field found ('latitude'/'longitude', 'lat/lon', 'y/x').<br>Use the Chart Options dialog to specify location fields.")

    def getChartContext(self, handlerId):
        diagTemplate = MapBoxBaseDisplay.__module__ + ":mapViewOptionsDialogBody.html"
        return (diagTemplate, {})
    
    def doRenderChart(self):
        mbtoken = self.options.get("mapboxtoken")
        if not mbtoken:
            return self.renderTemplate("noaccesstoken.html")
        else:
            self.response = requests.get("https://api.mapbox.com/tokens/v2?access_token=" + mbtoken)
            if (self.response.status_code == 200 and self.response.json()['code'] != 'TokenValid') \
                    or self.response.status_code != 200:
                return self.renderTemplate("tokenerror.html")

        body = self.renderMapView(mbtoken)
        if self.isStreaming:
            self.commId = str(uuid.uuid4())
            activesStreamingEntities[self.options.get("cell_id")] = self.entity

        return self.renderTemplate("iframesrcdoc.html", body=body, prefwidth=self.getPreferredOutputWidth(), prefheight=self.getPreferredOutputHeight())

    # override the default so that the non-numeric fields are collected as extra fields
    def getExtraFields(self):
        return self.getNonNumericValueFields()

    def renderMapView(self, mbtoken):

        # generate a working pandas data frame using the fields we need
        df = self.getWorkingPandasDataFrame()
        keyFields = self.getKeyFields()

        # geomType can be either 0: (Multi)Point, 1: (Multi)LineString, 2: (Multi)Polygon
        geomType = 0
        bins = []

        if len(keyFields)>0:
            if len(keyFields)==1:
                geomType = -1 #unknown as of yet
            else:
                lonFieldIdx = 0
                latFieldIdx = 1
                if keyFields[0] == self.getLatField(): 
                    lonFieldIdx = 1
                    latFieldIdx = 0
                min = [df[keyFields[lonFieldIdx]].min(), df[keyFields[latFieldIdx]].min()]
                max = [df[keyFields[lonFieldIdx]].max(), df[keyFields[latFieldIdx]].max()]
                self.options["mapBounds"] = json.dumps([min,max], default=defaultJSONEncoding)

        valueFields = self.getValueFields()
        
        #check if we have a preserveCols
        preserveCols = self.options.get("preserveCols", None)
        preserveCols = [a for a in preserveCols.split(",") if a not in keyFields and a not in valueFields] if preserveCols is not None else []

        # Transform the data into GeoJSON for use in the Mapbox client API
        allProps = valueFields + preserveCols + self.getExtraFields()
        features = []
        for rowidx, row in df.iterrows():
            feature = {'type':'Feature',
                        'properties':{},
                        'geometry':{'type':'Point',
                                    'coordinates':[]}}
            
            if geomType == 0:
                feature['geometry']['coordinates'] = [row[keyFields[lonFieldIdx]], row[keyFields[latFieldIdx]]]
            else:
                geomIdx = df.columns.get_loc(keyFields[0])+1
                feature['geometry'] = json.loads(row[geomIdx])
                
            for fld in allProps:
                feature['properties'][fld] = row[fld]
            features.append(feature)

        if len(features)>0:
            pygeojson = {'type':'FeatureCollection', 'features':features}
            self.options["mapData"] = json.dumps(pygeojson,default=defaultJSONEncoding)

            # Now let's figure out whether we have Line or Polygon data, if it wasn't already found to be Point
            if geomType != 1:
                if features[0]['geometry']['type'].endswith('LineString'):
                    geomType = 1
                elif features[0]['geometry']['type'].endswith('Polygon'):
                    geomType = 2
                else:
                    geomType = -1
                
            #### build up the map style

            # basic color
            paint = {}
            if geomType == 1:
                paint['line-color'] = '#ff0000'
                paint['line-width'] = 2
                if self.options.get("coloropacity"):
                    paint['line-opacity'] = float(self.options.get("coloropacity")) / 100
            elif geomType == 2:
                paint['fill-color'] = '#ff0000'
                paint['fill-opacity'] = 0.8
                if self.options.get("coloropacity"):
                    paint['fill-opacity'] = float(self.options.get("coloropacity")) / 100
            else:
                paint['circle-radius'] = 12
                paint['circle-color'] = '#ff0000'
                paint['circle-opacity'] = 0.25
                if self.options.get("coloropacity"):
                    paint['circle-opacity'] = float(self.options.get("coloropacity")) / 100
                if (self.options.get("kind") and self.options.get("kind").find("cluster") >= 0):
                    paint['circle-opacity'] = 1.0

            if len(valueFields) > 0:
                mapValueField = valueFields[0]
                self.options["mapValueField"] = mapValueField

            if not self.options.get("kind"): 
                self.options["kind"] = "choropleth-cluster"

            # if there's a numeric value field and type is not 'simple', paint the data as a choropleth map
            if self.options.get("kind") and self.options.get("kind").find("simple") < 0 and len(valueFields) > 0:
                # get value from the "Number of Bins" slider
                numBins = int(self.options.get("numbins", 5))

                # custom index, uses "Custom Base Color" hex value in options menu.
                customBaseColor = self.options.get("custombasecolor", "#ff0000")

                # custom index, uses "Secondary Custom Base Color" hex value in options menu.
                secondaryCustomBaseColor = self.options.get("custombasecolorsecondary", "#ff0000")

                # color options
                bincolors = []
                bincolors.append(self._getColorList('#ffffcc', '#253494', numBins)) #yellow to blue
                bincolors.append(self._getMonochromeLight('#de2d26', numBins)) # reds monochrome light
                bincolors.append(self._getMonochromeDark('#f7f7f7', numBins)) # grayscale monochrome dark
                bincolors.append(self._getColorList('#e66101', '#5e3c99', numBins)) # orange to purple
                bincolors.append(self._getColorList('#3E9E7A', '#0F091D', numBins)) # green to purple
                bincolors.append(self._getMonochromeLight(customBaseColor, numBins)) # custom monochrome light (saturation)
                bincolors.append(self._getMonochromeDark(customBaseColor, numBins)) # custom monochrome dark (value)
                bincolors.append(self._getColorList(customBaseColor, secondaryCustomBaseColor, numBins)) # custom range (value)

                bincolorsIdx = 0
                if self.options.get("colorrampname"):
                    if self.options.get("colorrampname") == "Light to Dark Red":
                        bincolorsIdx = 1
                    if self.options.get("colorrampname") == "Grayscale":
                        bincolorsIdx = 2
                    if self.options.get("colorrampname") == "Orange to Purple":
                        bincolorsIdx = 3
                    if self.options.get("colorrampname") == "Green to Purple":
                        bincolorsIdx = 4
                    if self.options.get("colorrampname") == "Custom Monochrome Light":
                        bincolorsIdx = 5
                    if self.options.get("colorrampname") == "Custom Monochrome Dark":
                        bincolorsIdx = 6
                    if self.options.get("colorrampname") == "Custom Color Range":
                        bincolorsIdx = 7

                # only use list of quantiles if it matches the number of bins
                if self.options.get("quantiles") and len(self.options.get("quantiles").split(",")) == numBins:
                    quantileFloats = [float(x) for x in self.options.get("quantiles").split(",")]
                    self.debug("Using quantileFloats: %s" % quantileFloats)
                    for i in range(numBins):
                        bins.append((df[valueFields[0]].quantile(quantileFloats[i]),bincolors[bincolorsIdx][i%len(bincolors[bincolorsIdx])]))
                else:
                    # default, equal-size bins based on numBins (if cannot find quantiles array in options)
                    self.debug("Using equal-size bins based on numBins: %s" % numBins)
                    for i in range(numBins):
                        bins.append((df[valueFields[0]].quantile(float(i)/(numBins-1.0)),bincolors[bincolorsIdx][i%len(bincolors[bincolorsIdx])]))

                if geomType == 1:
                    # paint['line-opacity'] = 0.65
                    paint['line-color'] = {"property":mapValueField}
                    paint['line-color']['stops'] = []
                    for bin in bins:
                        paint['line-color']['stops'].append([bin[0], bin[1]])
                elif geomType == 2:
                    paint['fill-color'] = {"property":mapValueField}
                    paint['fill-color']['stops'] = []
                    for bin in bins:
                        paint['fill-color']['stops'].append([bin[0], bin[1]])
                else:
                    # paint['circle-opacity'] = 0.65
                    paint['circle-color'] = {"property":mapValueField}
                    paint['circle-color']['stops'] = []
                    for bin in bins: 
                        paint['circle-color']['stops'].append([bin[0], bin[1]])
                    paint['circle-radius'] = 12


            self.options["mapStyle"] = json.dumps(paint,default=defaultJSONEncoding)
            
        w = self.getPreferredOutputWidth()
        h = self.getPreferredOutputHeight()

        # handle custom layers
        userlayers = []
        l = (ShellAccess,ShellAccess) 
        papp = self.options.get("nostore_pixieapp")
        if papp is not None and ShellAccess[papp] is not None:
            l = (ShellAccess[papp], dir(ShellAccess[papp]))
        for key in [a for a in l[1] if not callable(getattr(l[0], a)) and not a.startswith("_")]:
            v = getattr(l[0],key)            
            if isinstance(v, dict) and "maptype" in v and v["maptype"].lower() == "mapbox" and "source" in v and "type" in v["source"] and v["source"]["type"] == "geojson" and "id" in v and "data" in v["source"]:
                gj = geojson.loads(json.dumps(v["source"]["data"]))
                isvalid = True
                if hasattr(geojson, "is_valid"): # then we're using old version of geojson module
                    isvalid = geojson.is_valid(gj)["valid"] == "yes"
                    self.debug("IN hasattr(geojson,is_valid). Validity is "+str(isvalid))
                else: # we're using a newer version of geojson module
                    isvalid = gj.is_valid
                if isvalid:
                    userlayers.append(v)
                else:
                    self.debug("Invalid GeoJSON: {0}".format(str(v["source"]["data"])))
        self.debug("userlayers length: "+str(len(userlayers)))
        # end handle custom layers

        uniqueid = str(uuid.uuid4())[:8]
        return self.renderTemplate("mapView.html", bins=bins, userlayers=userlayers, prefwidth=w, prefheight=h, randomid=uniqueid)

    def isLatLonChart(self):
        llnames = ['lat','latitude','y','lon','long','longitude','x']
        isll = True;
        keyFields = self.getKeyFields()
        if ((keyFields is not None) and len(keyFields) == 2):
            for field in keyFields:
                if field.lower() not in llnames:
                    isll = False
        return isll;

    def getLatField(self):
        names = ['lat','latitude','y']
        keyFields = self.getKeyFields()
        if (keyFields is not None):
            for field in keyFields:
                if field.lower() in names:
                    return field
        return None;

    def getLonField(self):
        names = ['lon','long','longitude','x']
        keyFields = self.getKeyFields()
        if (keyFields is not None):
            for field in keyFields:
                if field.lower() in names:
                    return field
        return None;

    def _getDefaultKeyFields(self):
        # check for lat/long
        latLongFields = []
        for field in self.getFieldNames():
            if field.name.lower() == 'lat' or field.name.lower() == 'latitude' or field.name.lower() == 'y':
                latLongFields.append(field.name)
            elif field.name.lower() == 'lon' or field.name.lower() == 'long' or field.name.lower() == 'longitude' or field.name.lower() == 'x':
                latLongFields.append(field.name)
        if (len(latLongFields) == 2):
            return latLongFields
        # if we get here, look for an address field
        for field in self.getFieldNames():
            if field.name.lower() == 'address':
                return latLongFields.append(field.name)
        return []

    def _getColorList(self, color1, color2, nColors):
        colorList = []
        colorRange = Color(color1).range_to(Color(color2), nColors)
        for color in list(colorRange):
            colorList.append(color.hex_l)
        return colorList

    def _getMonochromeLight(self, customBaseColor, nColors):
        # HEX -> RGB
        customBaseColor = customBaseColor.lstrip("#")
        customBaseColorRGB = tuple(int(customBaseColor[i:i+2], 16) for i in (0, 2 ,4))
        # RGB -> HSV
        customBaseColorHSV = colorsys.rgb_to_hsv(customBaseColorRGB[0]/float(255), customBaseColorRGB[1]/float(255), customBaseColorRGB[2]/float(255))
        h, s, v = customBaseColorHSV
        increment = float(s) / float(nColors)
        customColorsArr = []
        for i in range(nColors):
            # HSV -> RGB
            customColorRGB = colorsys.hsv_to_rgb(h, s - (increment * (nColors - i)), v)
            # RGB -> HEX
            customColor = "#{:02x}{:02x}{:02x}".format( int(customColorRGB[0] * 255), int(customColorRGB[1] * 255), int(customColorRGB[2] * 255) )
            customColorsArr.append(customColor)
        return customColorsArr

    def _getMonochromeDark(self, customBaseColor, nColors):
        # HEX -> RGB
        customBaseColor = customBaseColor.lstrip("#")
        customBaseColorRGB = tuple(int(customBaseColor[i:i+2], 16) for i in (0, 2 ,4))
        # RGB -> HSV
        customBaseColorHSV = colorsys.rgb_to_hsv(customBaseColorRGB[0]/float(255), customBaseColorRGB[1]/float(255), customBaseColorRGB[2]/float(255))
        h, s, v = customBaseColorHSV
        increment = float(v) / float(nColors)
        customColorsArr = []
        for i in range(nColors):
            # HSV -> RGB
            customColorRGB = colorsys.hsv_to_rgb(h, s, v - (increment * i))
            # RGB -> HEX
            customColor = "#{:02x}{:02x}{:02x}".format( int(customColorRGB[0] * 255), int(customColorRGB[1] * 255), int(customColorRGB[2] * 255) )
            customColorsArr.append(customColor)
        return customColorsArr



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.chart.renderers import PixiedustRenderer
from pixiedust.display.chart.colorManager import Colors
from .matplotlibBaseDisplay import MatplotlibBaseDisplay
from pixiedust.utils import Logger
import numpy as np

@PixiedustRenderer(id="barChart")
@Logger()
class BarChartRenderer(MatplotlibBaseDisplay):

    def getNumFigures(self):
        return len(self.getValueFields()) if self.isSubplot() else 1

    def isSubplot(self):
        return self.options.get("charttype", "grouped") == "subplots"

    def getExtraFields(self):
        if not self.isSubplot() and len(self.getValueFields())>1:
            #no clusterby if we are grouped and multiValueFields
            return []
    
        clusterby = self.options.get("clusterby")
        return [clusterby] if clusterby is not None else []

    def getSubplotHSpace(self):
        return 0.5

    #Main rendering method
    def matplotlibRender(self, fig, ax):
        keyFields = self.getKeyFields()
        valueFields = self.getValueFields()
        stacked = self.options.get("charttype", "grouped") == "stacked"
        subplots = self.isSubplot()
        kind = "barh" if self.options.get("orientation", "vertical") == "horizontal" else "bar"
        clusterby = self.options.get("clusterby")

        if clusterby is not None and (subplots or len(valueFields)<=1):
            subplots = subplots if len(valueFields)==1 else False
            for j, valueField in enumerate(valueFields):
                pivot = self.getWorkingPandasDataFrame().pivot(
                    index=keyFields[0], columns=clusterby, values=valueField
                )
                pivot.index.name=keyFields[0]

                # sort order is lost after pivot
                sortby = self.options.get("sortby", None)
                if sortby:
                    if sortby == 'Keys ASC':
                        pivot = pivot.sort_values(keyFields, ascending=True)
                    elif sortby == 'Keys DESC':
                        pivot = pivot.sort_values(keyFields, ascending=False)
                    else:
                        pivot[valueFields[0]] = pivot.sum(axis=1)
                        if sortby == 'Values ASC':
                            pivot = pivot.sort_values(valueFields[0], ascending=True)
                        elif sortby == 'Values DESC':
                            pivot = pivot.sort_values(valueFields[0], ascending=False)
                        pivot.drop(valueFields[0], axis=1, inplace=True)

                thisAx = pivot.plot(kind=kind, stacked=stacked, ax=self.getAxItem(ax, j), sharex=True, legend=self.showLegend(), 
                    label=None if subplots else valueField, subplots=subplots,colormap = Colors.colormap)

                if isinstance(thisAx, (list, np.ndarray)):
                    if kind == 'barh':
                        for plotaxis in thisAx:
                            plotaxis.invert_yaxis()
                    if len(valueFields)==1 and subplots:
                        #resize the figure
                        figw = fig.get_size_inches()[0]
                        fig.set_size_inches( figw, (figw * 0.5)*min( len(thisAx), 10 ))
                elif kind == 'barh':
                    thisAx.invert_yaxis()

                return thisAx
        else:
            thisAx = self.getWorkingPandasDataFrame().plot(kind=kind, stacked=stacked, ax=ax, x=keyFields[0], legend=self.showLegend(), subplots=subplots,colormap = Colors.colormap)
            if kind == 'barh':
                if isinstance(thisAx, (list, np.ndarray)):
                    for plotaxis in thisAx:
                        plotaxis.invert_yaxis()
                else:
                    thisAx.invert_yaxis()

            if clusterby is not None:
                self.addMessage("Warning: 'Cluster By' ignored when you have multiple Value Fields but subplots option is not selected")


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2018
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
from pixiedust.display.app import *

@PixieApp
class TemplateTabbedApp():
    """Templated PixieApp that provide tab layout"""
    def setup(self):
        self.apps = []

    @route()
    def main_screen(self):
        if len(self.apps) == 0:
            return """
<h1>No app defined </h1>
<div>Try to define the apps in a setup method e.g.
<code>
def setup(self):
    self.apps=[{"title":"My app", "app_class": "my.package.MyApp"}]
</code>
"""
        
        return """
<ul class="nav nav-tabs" role="tablist">
{%for app in this.apps%}
    <li role="presentation" id="tab_{{prefix}}_{{loop.index}}" {%if loop.first%}class="active"{%endif%}>
        <a href="#content_{{prefix}}_{{loop.index}}" data-toggle="tab">{{app['title']}}</a>
    </li>
{%endfor%}
</ul>
<div class="tab-content">
{%for app in this.apps%}
    <div role="tabpanel" class="tab-pane {%if loop.first%}active{%endif%} no_loading_msg" 
        id="content_{{prefix}}_{{loop.index}}">
        <div pd_render_onload pd_app="{{app['app_class']}}"></div>
    </div>
{%endfor%}
</div>
<script>
$('#tab_{{prefix}}_1').click();
</script>        
"""


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
from collections import OrderedDict
from pixiedust.display.chart.renderers import PixiedustRenderer
from pixiedust.display.chart.renderers.baseChartDisplay import commonChartOptions
from pixiedust.utils import Logger
from .brunelBaseDisplay import BrunelBaseDisplay

@PixiedustRenderer(id="mapView")
@Logger()
class MapRenderer(BrunelBaseDisplay):

    def __init__(self, options, entity, dataHandler=None):
        super(MapRenderer, self).__init__(options, entity, dataHandler)
        self.types = OrderedDict([
            ('Map', {'handler': self.compute_for_map}),
            ('Heat Map', {'handler': self.compute_for_heatmap}),
            ('Tree Map', {'handler': self.compute_for_treemap}),
            ('Chord', {'handler': self.compute_for_chord}),
        ])

    @commonChartOptions
    def getChartOptions(self):
        return [
            {
                'name': 'brunelMapType',
                'description': 'Type',
                'refresh': False,
                'metadata': {
                    'type': 'dropdown',
                    'values': list(self.types.keys()),
                    'default': list(self.types.keys())[0]
                }
            }
        ]

    def compute_brunel_magic(self):
        map_type = self.options.get('brunelMapType', list(self.types.keys())[0])
        handler = self.types.get(map_type, list(self.types.values())[0]).get('handler')
        return handler()

    def compute_for_map(self):
        parts = ['map']
        parts.append("key({0}) label({0})".format(",".join(self.getKeyFields())))
        parts.append("color({})".format(",".join(self.getValueFields())))
        return parts

    def compute_for_heatmap(self):
        parts = []
        parts.append("x({})".format(",".join(self.getKeyFields())))
        parts.append("y({})".format(",".join(self.getValueFields())))
        parts.append("color(#count)")
        parts.append("style('symbol:rect; size:100%; stroke:none')")
        parts.append("sort(#count)")
        return parts
    
    def compute_for_treemap(self):
        parts = ['treemap']
        parts.append("x({0})".format(",".join(self.getKeyFields())))
        parts.append("color({0}) size({0}) label({0})".format(",".join(self.getValueFields())))
        return parts
    
    def compute_for_chord(self):
        parts = ['chord']
        parts.append("x({})".format(",".join(self.getKeyFields())))
        parts.append("y({})".format(",".join(self.getValueFields())))
        parts.append("color(#count)")
        parts.append("tooltip(#all)")
        return parts 


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

import ast
import inspect
from astunparse import unparse

def get_caller_text(frame):
    """
    Return the expression that calls the frame
    """
    def find_match_node(node):
        "Find a candidate ast node"
        match_node = None
        for chd in ast.iter_child_nodes(node):
            if getattr(chd, "lineno", 0) > frame.f_back.f_lineno:
                break
            match_node = node if isinstance(chd, ast.Name) and isinstance(node, ast.Call) else match_node
            match_node = find_match_node(chd) or match_node
        return match_node
    lines, _ = inspect.findsource(frame.f_back.f_code)
    match_node = find_match_node(ast.parse("".join(lines)))
    return unparse(match_node).strip().replace(', ', ',') if match_node is not None else None

def parse_function_call(expression):
    """
    Parse an expression looking for a function call
    return a dictionary with function name, args and kwargs
    e.g:
    {
        'func': 'display',
        'args': ['df'],
        'kwargs': {
            'aggregation': "'SUM'",
            'nostore_cw': "'1098'",
            'rowCount': "'500'"
        }
    }
    """
    class Walker(ast.NodeVisitor):
        "ast NodeVisition helper class"
        def __init__(self):
            self.results = {'func': None, 'args': [], 'kwargs': {}}

        def resolveName(self, node):
            if isinstance(node, ast.Name):
                self.results['func'] += node.id
            elif isinstance(node, ast.Attribute):
                self.resolveName(node.value)
                self.results['func'] += "." + node.attr

        def visit_Call(self, node):
            "Visit the first call function"
            if self.results['func'] is not None:
                return
            self.results['func'] = ""
            self.resolveName(node.func)
            self.results['args'] = [unparse(arg).strip() for arg in node.args]
            self.results['kwargs'] = {kw.arg:unparse(kw.value).strip().strip('\'"') for kw in node.keywords}

    walker = Walker()
    walker.visit(ast.parse(expression))
    return walker.results

def get_matches_lineno(code, fn_name):
    "Return a list of line number corresponding to the definition of function with the name fn_name"
    class Walker(ast.NodeVisitor):
        def __init__(self):
            self._hits = set()

        #pylint: disable=E0213,E1102    
        def onvisit(fn):
            def wrap(self, node):
                fn(self,node)
                super(Walker, self).generic_visit(node)
            return wrap
        
        @onvisit
        def visit_FunctionDef(self, node):
            if node.name == fn_name:
                self._hits.add(node)
            return node
        
        @property
        def hits(self):
            return list(sorted([n.lineno + 1 for n in self._hits]))
    
    walker = Walker()
    walker.visit(ast.parse(code))
    return walker.hits


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
from pixiedust.display.app import *
import sys
import os
import requests
import json
import nbformat
from pixiedust.utils.userPreferences import getUserPreference, setUserPreference, user_preferences
from pixiedust.utils import Logger
from pixiedust.apps.gateway import BaseGatewayApp
import warnings
import pkg_resources
from jupyter_client.manager import KernelManager
from IPython.core.getipython import get_ipython

import ast
class ImportsLookup(ast.NodeVisitor):
    def __init__(self):
        self.imports = set()

    #pylint: disable=E0213,E1102
    def onvisit(func):
        def wrap(self, node):
            ret_node = func(self, node)
            super(ImportsLookup, self).generic_visit(node)
            return ret_node
        return wrap

    def add_import(self, module_name):
        if not module_name in sys.builtin_module_names:
            try:
                module = __import__(module_name)
                ok_to_add = module.__package__ != ""
                if not ok_to_add and "site-packages" in module.__file__:
                    ok_to_add = True
                #check if egg-link (aka editable mode)
                if not ok_to_add:
                    for p in sys.path:
                        if os.path.isfile(os.path.join(p,module_name+".egg-link")):
                            ok_to_add = True

                if ok_to_add:
                    try:
                        pkg = pkg_resources.get_distribution(module_name)
                        version = pkg.parsed_version._version.release
                        self.imports.add( (module_name, version, self.get_egg_url(pkg, module_name)) )
                    except pkg_resources.DistributionNotFound:
                        pass
            except:
                print("Unknown import found {}".format(module_name))

    def get_egg_url(self, pkg, module_name):
        if requests.get("https://pypi.python.org/pypi/{}/json".format(module_name), timeout=5).status_code == 200:
            return None
        def try_location(base, name):
            ret = os.path.join(base,name)
            if not os.path.exists(ret):
                ret = os.path.join(base, name.replace("-", "_"))
            if not os.path.exists(ret):
                ret = os.path.join(base, name.replace("_", "-"))
            return ret
        
        location = pkg.location
        if not os.path.exists(os.path.join(location, "setup.py")):
            location = try_location(location, module_name)
        if not os.path.exists(os.path.join(location, "setup.py")):
            #try the github page
                pkg_info = os.path.join(pkg.location, "{}.egg-info".format(pkg.egg_name()), "PKG-INFO")
                if os.path.isfile(pkg_info):
                    with open(pkg_info) as fp:
                        for line in fp.readlines():
                            index = line.find(':')
                            if index>0:
                                key = line[0:index].strip()
                                if key.lower()=="home-page":
                                    return "git+{}#egg={}".format(line[index+1:].strip(), module_name)
    
    @onvisit
    def visit_ImportFrom(self, node):
        self.add_import(node.module.split(".")[0])
    
    @onvisit
    def visit_Import(self, node):
        for name in node.names:
            self.add_import(name.name)
        
    @onvisit
    def generic_visit(self, node):
        pass

@PixieApp
@Logger()
@user_preferences(security='token')
class PublishApp(BaseGatewayApp):
    """
    Publish a PixieApp as a web app
    """
    
    def setup(self):
        BaseGatewayApp.setup(self)
        self.kernel_spec = None
        self.kernel_name = None
        self.kernels = []
        self.lookup = None
        self.title = "Publish PixieApp to the web"
        self.tab_definitions = [{
            "title": "Basic Publish Options",
            "id": "options",
            "name": "Options",
            "contents": lambda: self.renderTemplate("publishBasicOptions.html")
        }, {
            "title": "Security",
            "id": "security",
            "name": "Security",
            "contents": lambda: self.renderTemplate("publishSecurityOptions.html")
        }, {
            "title": "Package dependencies",
            "id": "imports",
            "name": "Imports",
            "contents": lambda: self.renderTemplate("publishImportOptions.html")
        }, {
            "title": "Kernel Specification",
            "id": "kernelspec",
            "name": "Kernel Spec",
            "contents": lambda: self.renderTemplate("publishKernelSpecOptions.html")
        }]
        self.gateway_buttons = [{
            "title": "Publish",
            "options": ["server", "title", "icon", "sel_kernel"]
        }]
        self.set_gateway_server()

    def set_gateway_server(self, gateway_server = None):
        server = gateway_server.strip("/ ") if gateway_server is not None else self.server.strip("/ ")
        message = ""
        icon = "fa-check"
        try:
            #get the list of kernels
            response = requests.get("{}/stats/kernels".format(server), timeout=5)
            if response.status_code != requests.codes.ok:
                raise Exception(response.text)
            kernels_spec = response.json()
            if self.kernel_name and self.kernel_name in kernels_spec:
                for spec in kernels_spec.values(): spec['default'] = False
                kernels_spec[self.kernel_name]['default'] = True
            self.kernels = [(kernel, spec['default']) for kernel,spec in iteritems(kernels_spec)]

            message = "PixieGateway successfully validated. You can optionally pick a kernel in the Kernels tab"
        except Exception as exc:
            icon="fa-times"
            message = str(exc)
            self.error( "Unable to validate gateway server {} - {}".format(server, exc))


        if gateway_server is not None:
            print("""
                    <span style="bottom:0px;position:absolute">
                        <i style="font-size:2em" class="fa {}" aria-hidden="true"></i>
                        <span>{}</span>
                    </span>
                """.format(icon, message))

    def set_contents(self, contents):
        self.contents = json.loads(contents)
        self.contents['notebook']['metadata']['pixiedust'] = {}
        kernel_spec = self.contents['notebook']['metadata']['kernelspec']
        self.kernel_name=kernel_spec['name']
        km = KernelManager(kernel_name=kernel_spec['name'])
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.kernel_spec = json.dumps(km.kernel_spec.to_dict(), indent=4, sort_keys=True)

    @route(gateway_server="*")
    def publish(self, gateway_server, gateway_title, gateway_icon, gateway_sel_kernel):
        self.server = gateway_server.strip("/ ")
        self.contents['notebook']['metadata']['pixiedust'].update({
            "title":gateway_title, "icon":gateway_icon, "security": self.security
        })
        if gateway_sel_kernel != "Default":
            self.contents['notebook']['metadata']['pixiedust'].update({"kernel":gateway_sel_kernel})
        self.compute_imports()
        setUserPreference("pixie_gateway_server", gateway_server)
        response = requests.post(
            "{}/publish/{}".format(self.server, self.contents['name']), 
            json = self.contents['notebook']
        )
        if response.status_code == requests.codes.ok:
            self.pixieapp_model = response.json()
            return """
<style type="text/css">
.publish{
    font-size: larger;
    margin-left: 30px;
}
.publish .logmessages{
}
.publish .logmessage{
    color: darkblue;
}
.publish .summary{
    font-size: xx-large;
    text-align: center;
}
</style>
<div class="publish">
    <div class="logmessages">
        {%for message in this.pixieapp_model['log']%}
        <div class="logmessage">
            {{message}}
        </div>
        {%endfor%}
    </div>
    <div class="summary">
        <div>Notebook Successfully published</div>
        <div>
            <a href="{{this.pixieapp_model['url']}}" target="blank">
                {{this.contents['name']}}
            </a>
        </div>
    </div>
</div>
            """
        
        return "<div>An Error occured while publishing this notebook: {}".format(response.text)

    
    def ast_parse(self, code):
        try:
            return ast.parse(code)
        except SyntaxError:
            #transform the code first to handle notebook syntactic sugar like magic and system
            return ast.parse(
                get_ipython().input_transformer_manager.transform_cell(code)
            )

    def compute_imports(self):
        if self.lookup is None:
            notebook = nbformat.from_dict(self.contents['notebook'])
            code = ""
            for cell in notebook.cells:
                if cell.cell_type == "code":
                    code += "\n" + cell.source                
            self.lookup = ImportsLookup()
            self.lookup.visit(ast.parse(self.ast_parse(code)))
            self.contents['notebook']['metadata']['pixiedust'].update({
                "imports": {p[0]:{"version":p[1],"install":p[2]} for p in self.lookup.imports}
            })
        
    @route(importTable="*")
    def imports(self):
        self.compute_imports()
        return """
<table class="table">
    <thead>
        <tr>
            <th>Package</th>
            <th>Version</th>
            <th>Install</th>
        </tr>
    </thead>
    <tbody>
        {% for import in this.lookup.imports%}
        <tr>
            <td>{{import[0]}}</td>
            <td>{{import[1]}}</td>
            <td>{{import[2] or "PyPi"}}</td>
        </tr>
        {%endfor%}
    </tbody>
</table>
        """
    
    @route(showKernelSpec="*")
    def show_kernel_spec(self):
        return "<pre>{{this.kernel_spec}}</pre>"



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2018
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.app import route
from pixiedust.utils import Logger

@Logger()
class AggregationSelector(object):
    @route(widget="pdAggregationSelector")
    def chart_option_aggregation_widget(self, optid, aggregation):
        self.aggOpts = self.aggregation_options()
        self.aggOpts.sort()
        if not aggregation:
            aggregation = self.aggregation_default()
        self.selectedAgg = aggregation
        return """
<div class="form-group">
    <label class="field">Aggregation:</label>
    <select id="chartoption{{optid}}{{prefix}}" class="form-control" name="{{optid}}" style="margin-left: 0px;"
        onchange="$(this).find('> option:selected').trigger('click')"
        pd_script="self.options_callback('{{optid}}','$val(chartoption{{optid}}{{prefix}})')">
        {% for aggOpt in this.aggOpts %}
        <option value="{{aggOpt}}" {{'selected' if this.selectedAgg == aggOpt}}>{{aggOpt}}</option>
        {% endfor %}
    </select>
</div>
"""

    def aggregation_options(self):
        return ["SUM","AVG","MIN","MAX","COUNT"]

    def aggregation_default(self):
        return self.get_renderer.getDefaultAggregation(self.parsed_command['kwargs']['handlerId'])




# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
from pixiedust.display.app import *
from repository.mlrepositoryclient import MLRepositoryClient
from repository.mlrepositoryartifact import MLRepositoryArtifact
from pixiedust.utils.shellAccess import ShellAccess

@PixieApp
class WatsonMLApp():    
    def initWatsonML(self):
        credentials = self.getPixieAppEntity()
        if credentials is None:
            return "<div>You must provide credentials to your Watson ML Service</div>"
        self.ml_repository_client = MLRepositoryClient(credentials['service_path'])
        self.ml_repository_client.authorize(credentials['username'], credentials['password'])
        
    def downloadModel(self, modelId):
        ShellAccess["mlModel"] = self.ml_repository_client.models.get(modelId).model_instance()
        print("Model successfully downloaded in variable mlModel")
        
    @route()
    def startScreen(self):
        message = self.initWatsonML()
        if message is not None:
            return message

        return """
<style>
.pd_tooltip {
    position: relative;
    display: inline-block;
    border-bottom: 1px dotted black;
}

.pd_tooltip .pd_tooltiptext {
    visibility: hidden;
    background-color: white;
    color: black;
    border: 1px solid black;
    border-radius: 6px;
    padding: 5px 0;
    position: absolute;
    z-index: 1;
    bottom: 125%;
    margin-left: -360px;
    opacity: 0;
    transition: opacity 1s;
}

.pd_tooltip .pd_tooltiptext::after {
    content: "";
    position: absolute;
    top: 100%;
    left: 50%;
    margin-left: -5px;
    border-width: 5px;
    border-style: solid;
    border-color: #555 transparent transparent transparent;
}

.pd_tooltip:hover .pd_tooltiptext {
    visibility: visible;
    opacity: 1;
}
</style>
<div class="well">
    <span style="font-size:x-large">List of Watson ML models</span>
    <button style="float:right" type="submit" class="btn btn-primary">
        Upload a new Model
    </button>
</div>

{%for model in this.ml_repository_client.models.all()%}
    {%if loop.first or ((loop.index % 4) == 1)%}
<div class="row">
    <div class="col-sm-2"/>
    {%endif%}
    <div class="col-sm-2" style="border: 1px solid lightblue;margin: 10px;border-radius: 25px;cursor:pointer;
        min-height: 150px;background-color:#e6eeff;display: flex;align-items: center;">
        <table style="margin-left:15px;width:90%">
            <tr>
                <td style="padding:10px"><b>Model:</b></td>
                <td>{{model.name}}</td>
            </tr>
            <tr>
                <td style="padding:10px"><b>Id:</b></td>
                <td>{{model.uid}}</td>
            </tr>
            <tr>
                <td style="padding:10px"><b>Creation Time:</b></td>
                <td>{{model.meta.prop("creationTime")}}</td>
            </tr>
            <tr>
                <td colspan="2" style="text-align:center">
                    <div class="btn-group">
                        <button type="button" class="btn btn-default" pd_script="self.removeModel('{{model.uid}}')">
                            <i class="fa fa-trash"/>
                        </button>
                        <button type="button" class="btn btn-default" pd_target="message{{loop.index}}{{prefix}}" 
                            pd_script="self.downloadModel('{{model.uid}}')">
                            <i class="fa fa-download"/>
                        </button>
                        <button type="button" class="pd_tooltip btn btn-default">
                            <i class="fa fa-info"/>
                            <span class="pd_tooltiptext">
                                <table>
                                {% for item in model.meta.available_props() if (item != "inputDataSchema" and item !="trainingDataSchema")%}
                                    <tr>
                                        <td><b>{{item}}</b></td>
                                        <td>{{model.meta.prop(item)}}</td>
                                    </tr>
                                {% endfor %}
                                </table>
                            </span>
                        </button>                        
                    </div>
                </td>
            </tr>
            <tr>
                <td colspan="2">
                    <div id="message{{loop.index}}{{prefix}}"/>
                </td>
            </tr>
        </table>
    </div>
    {%if loop.last or ((loop.index % 4) == 0)%}
    <div class="col-sm-2"/>
</div>
    {%endif%}
{%endfor%}
"""


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2018
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.app import route
from pixiedust.utils import Logger

@Logger()
class BinQuantiles(object):
    @route(widget="pdBinQuantiles")
    def chart_option_binranges_widget(self, optid, numbins, quantiles):
        htmlArr = []
        htmlArr.append("""
<script type="text/javascript">
// need to return list back as a string in the form "0.0,0.25,0.50,0.75,1.0"
function getQuantilesStr() {
    var quantiles = [];
    $('.quantile-val').each(function() {
        quantiles.push($(this).val());
    });
    return quantiles.join(',');
}
// check if sorted and all in between 0.0 and 1.0
function validateQuantiles(quantilesStr) {
    var quantiles = quantilesStr.split(',').map(parseFloat);
    var isValid = true;
    var prev = quantiles[0];
    isValid = prev >= 0.0 && prev <= 1.0;
    for (var i = 1; i < quantiles.length; i++) {
        var curr = quantiles[i];
        if (prev < curr && curr >= 0.0 && curr <= 1.0) {
            prev = curr;
        } else {
            console.log('some data is invalid');
            isValid = false;
            break;
        }
    }
    return isValid;
}
function updateQuantiles() {
    var quantilesStr = getQuantilesStr();
    chartOptionsOKStatus{{prefix}}(!validateQuantiles(quantilesStr));
    return quantilesStr;
}
updateQuantiles();
</script>
<p>Edit bin quantiles.  Values should be between [0.0, 1.0].  Bins should be in ascending order.</p><br/>
        """)
        quantileVals = quantiles.split(",")
        if len(quantileVals) != int(numbins):
            quantileVals = self.generateQuantiles(int(numbins)).split(",")
        for i in range(int(numbins)):
            htmlArr.append("""
<div class="form-group">
    <label for="quantile_{0}_{{{{optid}}}}{{{{prefix}}}}">Quantile {0} Percentage:</label>
    <input id="quantile_{0}_{{{{optid}}}}{{{{prefix}}}}" class="quantile-val form-control" name="quantile_{0}" type="number" min="0.0" max="1.0" step="0.01" value="{1}"
        pd_script="self.options_callback('quantiles','$val(updateQuantiles)')" onkeyup="updateQuantiles()">
</div>
        """.format(i+1, quantileVals[i]))
        return "".join(htmlArr)



'''
-------------------------------------------------------------------------------

Copyright IBM Corp. 2017

Licensed under the Apache License, Version 2.0 (the 'License');

you may not use this file except in compliance with the License.

You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software

distributed under the License is distributed on an 'AS IS' BASIS,

WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.

See the License for the specific language governing permissions and

limitations under the License.

-------------------------------------------------------------------------------
'''

# MongoDB Browser for PixieDust
# Author: Nick Kasten
# Summer 2017

import ssl
import re
import pandas as pd
from pixiedust.display.app import *
try:
    import pymongo
    from bson.json_util import dumps
    display_dependency_instructions = False
except NameError:
    display_dependency_instructions = True

@PixieApp
class mongo_db_browser:
    #-------------------------- UTILITY METHODS --------------------------#
    def create_client(self, flag=None):
        try:
            client = pymongo.MongoClient(self.mongo_uri + "&ssl_cert_reqs=CERT_NONE")
        except pymongo.errors.PyMongoError:
            client = None
            self.exception("Error connecting to Mongo DB.")
        return client
        
    def get_docs(self, start=0, end=None):
        return list(self.client[self.db][self.collection].find()[start:end])

    def format_raw_query(self):
        split_query = [x.lstrip() for x in re.sub(r'(?:&quot;)|(?:&apos;)', '"', self.raw_query).split('\n') if len(x) > 6 and len(x.strip()) != 0]
        formatted_sort_list = []
        for item in eval(split_query[2][8:]):
            formatted_sort_list += [(key, self.convert_dir(val)) for (key, val) in item.items()]
        return [eval(split_query[0][12:-1]), eval(split_query[1][14:-1]), formatted_sort_list]

    def get_query_docs(self, start=0, end=None):
        return list(self.client[self.db][self.collection].find(self.formatted_query[0], self.formatted_query[1])[start:end])

    def convert_dir(self, dir_string):
        if dir_string == 'ascending' or dir_string == 'asc':
            return 1
        elif dir_string == 'descending' or dir_string == 'desc':
            return -1

    def get_dataframe(self):
        return self.output_dataframe

    def go_back(self):
        if int(self.doc_index) >= self.page_size:
            return self.doc_index-self.page_size
        else:
            return self.doc_index
            
    def go_fwd(self):
        if int(self.doc_index) <= self.doc_count-self.page_size:
            return self.doc_index+self.page_size
        else:
            return self.doc_index
    #-------------------------- DYNAMIC CONTENT METHODS --------------------------#
    def create_list_head(self, item_type):
        if item_type == "db":
            title = """Select a DB to see a list of its collections:"""
        elif item_type == "collection":
            title = """Select a Collection to see available documents:"""
        return """
        <style>
            #button_list {
                list-style: none;
                padding: 0;
                margin: 0;
            }
            i {
                margin-right: 10px;
            }
            li {
                margin-bottom: 10px;
            }
        </style>
        <link href="https://maxcdn.bootstrapcdn.com/font-awesome/4.7.0/css/font-awesome.min.css" rel="stylesheet" integrity="sha384-wvfXpqpZZVQGK6TAh5PVlGOfQNHSoD2xbE+QkPxCAFlNEevoEH3Sl0sibVcOQVnN" crossorigin="anonymous">
        <div class="row text-center">
            <div class="jumbotron">
                <h2>""" + title + """</h2>
            </div>"""

    def create_button_list(self, item_type, button_names):
        if item_type == "db":
            icon_name = "fa-database"
            script = """'self.db="{0}"'"""
            view = "collection_list"
        elif item_type == "collection":
            icon_name = "fa-files-o"
            script = """'self.collection="{0}"'"""
            view = "doc_preview"

        button_list = """<ul id="button_list">"""
        for name in button_names:
            button_list += """
            <li>
                <button 
                    id=""" + name + """
                    class="btn btn-primary"
                    pd_script=""" + script.format(name) + """
                    pd_options="view=""" + view +"""">
                    <i class="fa """ + icon_name + """" aria-hidden="true"></i> 
                    """ + name + """
                </button>
            </li>"""
        button_list += """
            <li>
                <button 
                    pd_options="view="
                    class="btn btn-default">Back
                </button>
            </li>
        </ul>"""
        return button_list

    def create_side_navbar(self):
        return """
        <div id="coll_controls" class="col-sm-2">
            <button 
                style="margin-bottom: 10px;"
                pd_options="view=collection_list"
                class="btn btn-default">Back
            </button>
            <h3>{{this.collection}}</h3>
            <button 
                style="margin-top: 10px;"
                pd_options="view=doc_viewer"
                pd_script="self.all_docs=self.get_docs()\nself.doc_index=0\nself.all_doc_view='true'\nself.dataframe_generated='false'"
                class="btn btn-primary">View All Documents
            </button>
            <button 
                style="margin-top: 10px;"
                pd_options="view=create_query"
                pd_script="self.all_doc_view='false'\nself.dataframe_generated='false'"
                class="btn btn-primary">Query Documents
            </button>
        </div>"""

    def create_doc_viewer_head(self):
        return """
        <style>
            #doc_container {
                padding: 0px 10px 0px 30px;
            }
            #coll_controls {
                padding: 10px;
            }
            #outer_row {
                margin: 10px;
                border: 1px solid #ddd;
                border-radius: 4px 4px 0px 0px;
            }  
            #viewer_heading {
                padding: 0px 0px 10px 10px;
            }
            #mid_row {
                padding: 15px;
                margin-bottom: 0px;
                background-color: #f5f5f5;
            }
            #doc_navbar {
                margin-top: 10px;
            }
            
            #doc_navbar span {
                padding-left: 5px;
            }
            textarea {
                margin-bottom: 15px; 
                padding: 10px; 
                font-family: monospace;
                font-size: 9pt; 
                background-color: white; 
                max-height: 200px; 
                white-space: pre; 
                overflow: scroll; 
                border-style: solid; 
                border-radius: 2px; 
                border-width: 1px; 
                border-color: #aaa;
            }
            .hiddenWell {
                margin: 0px 20px 0px 20px; 
                padding: 10px; 
                font-family: monospace;
                font-size: 9pt; 
                background-color: white; 
                max-height: 200px; 
                white-space: pre; 
                overflow: scroll; 
                border-style: solid; 
                border-radius: 2px; 
                border-width: 1px; 
                border-color: #aaa;
            }            
            .doc_element {
                margin-top: 10px;
                font-size: 9pt;
            }
            .doc_element a {
                text-decoration: none !important;
            }
        </style>"""

    def create_doc_viewer(self):
        page_title = 'All Documents' if self.all_doc_view == 'true' else 'Query Results'
        list_head = """<div class="row">""" + self.create_side_navbar() + """ 
            <div id="doc_container" class="col-sm-10">
                <div id="outer_row" class="row">
                    <h4 id="viewer_heading">""" + page_title + """</h4>
                    <div id="mid_row" class="row">
                        <div id="inner_row" class="row">
                            <div id="doc_target{{prefix}}">"""
        current_page_range = [x for x in self.all_docs[self.doc_index: self.doc_index+self.page_size]]
        doc_list = self.create_doc_elements(current_page_range)
        list_tail = self.create_doc_navbar() + """
                            </div>
                        </div>
                    </div>
                </div>
            </div>
        </div>"""
        return list_head + doc_list + list_tail

    def create_doc_navbar(self):
        output = """
        <div id="doc_navbar">
            {% if this.doc_index >= this.page_size %}
                <button 
                    pd_options="view=doc_viewer" 
                    pd_script="self.doc_index=self.go_back()" 
                    type="button" 
                    class="btn btn-primary btn-sm control_btn">Prev
                </button>
            {% endif %}
            {% if this.doc_index <= this.doc_count - this.page_size %}
                <button 
                    pd_options="view=doc_viewer" 
                    pd_script="self.doc_index=self.go_fwd()" 
                    type="button" 
                    class="btn btn-primary btn-sm control_btn">Next
                </button>
            {% endif %}"""
        output += """
            <span><b>{0} to {1} of {2}</b></span>
            <hr style="border: 1px solid #ddd;">""".format(self.doc_index+1, self.doc_index+self.page_size, self.doc_count)
        output += """<h4>Access dataframe with .get_dataframe() method.</h4>""" if self.dataframe_generated == 'true' else """
            <button 
                pd_options="view=doc_viewer"
                pd_script="self.output_dataframe=pd.DataFrame(self.all_docs)\nself.dataframe_generated='true'" 
                type="button" 
                class="btn btn-info btn control_btn">Generate Dataframe
            </button>
        </div>"""
        return output

    def create_doc_elements(self, doc_list):
        output = ""
        count = 0
        for doc in doc_list:
            count += 1
            doc_dict = [[x, y] for (x, y) in zip(doc.keys(), doc.values())]
            preview_lines = 4 if (len(doc_dict) >= 4) else len(doc_dict)
            output += """
            <div class="doc_element">
                <a href="#doc""" + str(count) + """-{{prefix}}" data-toggle="collapse"> """ + """<pre> {0} {{</pre>""".format(doc['_id'])
            for i in list(range(1, preview_lines)):
                if len(str(doc_dict[i][1])) > 20:
                    doc_dict[i][1] = str(doc_dict[i][1])[0:20] + '...'
                output += """<pre>  "{0}": {1}</pre>""".format(doc_dict[i][0], doc_dict[i][1])
            output += """<pre>  ...</pre></a>
                <div id="doc""" + str(count) + """-{{prefix}}" class="collapse hiddenWell">""" + dumps(doc, indent=2, sort_keys="true") + """
                </div><a href="#doc""" + str(count) + """-{{prefix}}" data-toggle="collapse">""" +  """<pre>}</pre></a>
            </div>"""
        return output
    #-------------------------- ROUTES/VIEWS --------------------------#
    @route()
    def default_route(self):
        self.mongo_uri = ""
        self.page_size = 5

        if display_dependency_instructions == True:
            output = """<div class="row text-center" style="font-size: 12pt;">Please install <b>pymongo</b> by running the following command: <code>!pip install pymongo</code></div>"""
        else:
            output = """
        <div class="row">
            <h3>Enter your MongoDB URI to list available DBs:</h3>
        </div>
        <div class="row">
            <div class="form-group">
                <input 
                    id="mongo_uri"
                    type="text" 
                    class="form-control" 
                    rows="1" 
                />
                <button 
                    type="submit" 
                    style="margin-top: 5px;"
                    pd_options="view=db_list"
                    pd_script="self.mongo_uri='$val(mongo_uri)'" 
                    class="btn btn-success">Connect
                </button>
            </div>
        </div>"""
        return output

    @route(view="db_list")
    def db_list(self):
        self.client = self.create_client('no_ssl_cert')
        return self.create_list_head("db") + self.create_button_list("db", self.client.database_names()) + """</div>"""

    @route(view="collection_list")
    def collection_list(self):
        return self.create_list_head("collection") + self.create_button_list("collection", self.client[self.db].collection_names()) + """</div>"""

    @route(view='doc_preview')
    def doc_preview(self):
        return """<div class="row">""" + self.create_side_navbar() + """</div>"""

    @route(view='doc_viewer')
    def doc_viewer(self):
        self.doc_count = len(self.all_docs)
        return self.create_doc_viewer_head() + self.create_doc_viewer()

    @route(view='create_query')
    def create_query(self):
        self.query = """
{
    "selector": {"_id": {"$exists": "true"}},
    "projection": ["_id"],
    "sort": [{"_id": "ascending"}]
}"""
        return self.create_doc_viewer_head() + """
        <div class="row"> """ + self.create_side_navbar() + """
            <div id="doc_container" class="col-sm-10">
                    <div id="outer_row" class="row">
                        <h4 id="viewer_heading">Query</h4>
                        <div id="mid_row" class="row">
                            <div id="inner_row" class="row">
                                <div class="form-group">
                                    <textarea rows="6" class="form-control" id="query{{prefix}}">""" + self.query + """
                                    </textarea>  
                                    <button 
                                        type="submit"
                                        pd_options="view=doc_viewer"
                                        pd_script='self.raw_query="$val(query{{prefix}})"\nself.formatted_query=self.format_raw_query()\nself.doc_index=0\nself.all_docs=self.get_query_docs()'
                                        class="btn ">Go
                                    </button>
                                </div>
                            </div>
                        </div>
                    </div>
                </div>
            </div>
        </div>"""


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.chart.renderers import PixiedustRenderer
from .bokehBaseDisplay import BokehBaseDisplay
from pixiedust.utils import Logger
from bokeh.plotting import figure
from bokeh.models import HoverTool
import sys


@PixiedustRenderer(id="lineChart")
@Logger()
class BKLineChartRenderer(BokehBaseDisplay):
    def isSubplot(self):
        return self.options.get("lineChartType", "grouped") == "subplots"

    def getExtraFields(self):
        if not self.isSubplot() and len(self.getValueFields())>1:
            #no clusterby if we are grouped and multiValueFields
            return []
    
        clusterby = self.options.get("clusterby")
        return [clusterby] if clusterby is not None else []

    def createBokehChart(self):
        keyFields = self.getKeyFields()
        valueFields = self.getValueFields()
        clusterby = self.options.get("clusterby")
        stacked = self.options.get("charttype", "subplots") == "grouped"
        multivalue =  len(valueFields) > 1

        def lineChart(df, xlabel, vFields, color=None, clustered=None, title=None):
            ylabel = ','.join(v for v in vFields)
            x = list(df[xlabel].values)
            if df[xlabel].dtype == object:
                p = figure(y_axis_label=ylabel, x_axis_label=xlabel, title=title, x_range=x, **self.get_common_figure_options())
            else:
                p = figure(y_axis_label=ylabel, x_axis_label=xlabel, title=title, **self.get_common_figure_options())

            if clustered is not None:
                colors = self.colorPalette(len(df[clustered].unique())) if color is None else color
                df[clustered] = df[clustered].astype(str)

                for j,c in enumerate(list(df[clustered].unique())) if clustered else enumerate([None]):
                    df2 = df[df[clustered] == c] if c else df
                    df2 = df2.drop(clustered, axis=1) if c else df2

                    for i,v in enumerate(vFields):
                        y = list(df2[v].values)
                        l = v if self.isSubplot() else c
                        p.line(x, y, line_width=2, color=colors[i] if self.isSubplot() else colors[j], legend=l if self.showLegend() else None)
            else:
                colors = self.colorPalette(len(vFields)) if color is None else color
                

                for i,v in enumerate(vFields):
                    y = list(df[v].values)
                    p.line(x, y, line_width=2, color=colors[i], legend=v if self.showLegend() else None)

            p.legend.location = "top_left"

            hover = HoverTool()
            hover.tooltips = [(xlabel, '@x'), (ylabel, '@y{0.00}'), ('x', '$x'), ('y', '$y')]
            p.add_tools(hover)

            return p

        charts = []
        wpdf = self.getWorkingPandasDataFrame().copy()
        xlabel = keyFields[0] # ','.join(k for k in keyFields)

        for index, row in wpdf.iterrows():
            for k in keyFields:
                if isinstance(row[k], str if sys.version >= '3' else basestring):
                    row[k] = row[k].replace(':', '.')
            wpdf.loc[index] = row

        wpdf.sort_values(xlabel, ascending=[True], inplace=True)

        if self.isSubplot():
            colors = self.colorPalette(len(valueFields)) if len(valueFields) > 1 else self.colorPalette(len(valueFields) * (len(wpdf[clusterby].unique()) if clusterby else len(valueFields)))
            if clusterby:
                for j,c in enumerate(list(wpdf[clusterby].unique())):
                    wpdf2 = wpdf[wpdf[clusterby] == c]
                    charts.append(lineChart(wpdf2, xlabel, valueFields, clustered=clusterby, color=colors, title="{} = {}".format(clusterby, c)))
            else:
                for i, vField in enumerate(valueFields):
                    charts.append(lineChart(wpdf, xlabel, [valueFields[i]], color=colors[i:]))

        elif clusterby and not multivalue:
            charts.append(lineChart(wpdf, xlabel, valueFields, clustered=clusterby))

        else:
            charts.append(lineChart(wpdf, xlabel, valueFields))
            if clusterby:
                self.addMessage("Warning: 'Cluster By' ignored when grouped option with multiple Value Fields is selected")

        return charts



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2018
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from abc import ABCMeta
from six import with_metaclass
from pixiedust.display.chart.renderers import PixiedustRenderer
from pixiedust.display.chart.renderers.baseChartDisplay import BaseChartDisplay, commonChartOptions

@PixiedustRenderer(rendererId="table")
class TableBaseDisplay(with_metaclass(ABCMeta, BaseChartDisplay)):
    
    def get_options_dialog_pixieapp(self):
            """
            Return the fully qualified path to a PixieApp used to display the dialog options
            PixieApp must inherit from pixiedust.display.chart.options.baseOptions.BaseOptions
            """
            return "pixiedust.display.chart.renderers.table.tableOptions.TableOptions"

    ## overrides empty method in baseChartDisplay
    def getExtraFields(self):
        fieldNames = self.getFieldNames()
        if len(fieldNames) == 0:
            return []
        tableFields = []
        tableFieldStr = self.options.get("tableFields")
        if tableFieldStr is None:
            return fieldNames
        else:
            tableFields = tableFieldStr.split(",")
            tableFields = [val for val in tableFields if val in fieldNames]
        return tableFields

    def canRenderChart(self):
        return (True, None)

    def isTableRenderer(self):
        return True
    
    def supportsKeyFields(self, handlerId):
        return False

    def supportsKeyFieldLabels(self, handlerId):
        return False

    def supportsLegend(self, handlerId):
        return False

    def supportsAggregation(self, handlerId):
        return False

    @commonChartOptions
    def getChartOptions(self):
        return [
            {
                'name': 'table_noschema',
                'description': 'Hide Schema',
                'metadata': {
                    'type': 'checkbox',
                    'default': "false"
                }
            },
            {
                'name': 'table_nosearch',
                'description': 'Hide Search',
                'metadata': {
                    'type': 'checkbox',
                    'default': "false"
                }
            },
            {
                'name': 'table_nocount',
                'description': 'Hide Row Count',
                'metadata': {
                    'type': 'checkbox',
                    'default': "false"
                }
            },
            {
                'name': 'table_showrows',
                'description': 'Show Rows',
                'tooltip': 'Warning: this filter only applies to the sampled data and not the original data set',
                'metadata': {
                    'type': 'dropdown',
                    'values': ['All', 'Missing values', 'Not missing values'],
                    'default': "All"
                }
            }
        ]
    
